import wrapt
import abc
import copy
import json

from wafec_wrapt_custom.utility import fullname, starts_with

__all__ = [
    'WafecDefaultProxy',
    'create_proxy',
    'WafecDefaultProxyIterator'
]


_simple_types = (int, float, str, bool, complex)
_json_default_encoder_default = json._default_encoder.default
_white_list = ['nova.']


def _is_allowed(x):
    if not isinstance(x, wrapt.ObjectProxy):
        if isinstance(x, dict) or isinstance(x, list) or callable(x) or\
           type(x).__name__ == 'dict_items' or isinstance(x, tuple):
            return True
    if x is not None and type(x) not in _simple_types:
        if starts_with(_white_list, x):
            return True
    return False


def _resolve_json_encoder():
    def default(cls, o):
        if isinstance(o, WafecDefaultProxy):
            return o.__wrapped__
        raise TypeError('')
    json._default_encoder.default = default.__get__(json._default_encoder, _json_default_encoder_default)


def create_proxy(x):
    if _is_allowed(x):
        return WafecDefaultProxy(x)
    return x


class WafecDefaultProxy(wrapt.ObjectProxy, abc.ABC):
    def __init__(self, wrapped):
        super(WafecDefaultProxy, self).__init__(wrapped)
        _resolve_json_encoder()

    def __copy__(self):
        result = copy.copy(self.__wrapped__)
        return create_proxy(result)

    def __deepcopy__(self, memodict={}):
        result = copy.deepcopy(self.__wrapped__, memodict)
        return create_proxy(result)

    def __getitem__(self, item):
        result = self.__wrapped__[item]
        return create_proxy(result)

    def __getattr__(self, item):
        result = getattr(self.__wrapped__, item)
        return create_proxy(result)

    def __call__(self, *args, **kwargs):
        result = self.__wrapped__(*args, **kwargs)
        return create_proxy(result)

    def __iter__(self):
        return WafecDefaultProxyIterator(self.__wrapped__)

    def __reduce__(self):
        result = self.__wrapped__.__reduce__()
        return result

    def __reduce_ex__(self, protocol):
        result = self.__wrapped__.__reduce_ex__(protocol)
        return result


class WafecDefaultProxyIterator:
    def __init__(self, wrapped):
        self.wrapped = wrapped
        self.iterator = iter(wrapped)

    def __iter__(self):
        return self

    def __next__(self):
        result = next(self.iterator)
        return create_proxy(result)



import unittest

from wafec_wrapt_custom.wrappers import WrapperTest, WrapperData


class Aux(object):
    def __init__(self):
        self.item1 = Item(1)
        self.item2 = Item(2)

    def __repr__(self):
        return '<Aux(item1={}, item2={})>'.format(self.item1.value, self.item2.value)


class Item(object):
    def __init__(self, value):
        self.value = value


class TestWrapperTest(unittest.TestCase):
    def test_hello(self):
        wrapped = Aux()
        wrapper = WrapperTest(wrapped)
        self.assertTrue(isinstance(wrapper, Aux))
        self.assertTrue(isinstance(wrapper, WrapperTest))
        self.assertTrue(isinstance(wrapper.item1, WrapperTest))
        item3 = Item(3)
        wrapper.item2 = item3
        self.assertTrue(isinstance(wrapper.item2, WrapperTest))
        self.assertEqual(3, wrapper.item2.value)
        self.assertTrue(isinstance(wrapper.item2.value, int))

    def test_list(self):
        wrapped = [Aux(), Aux(), Aux()]
        wrapper = WrapperTest(wrapped)
        for item in wrapper:
            self.assertTrue(isinstance(item, WrapperTest))

    def test_dict(self):
        wrapped = {'item1': Aux(), 'item2': Aux()}
        wrapper = WrapperTest(wrapped)
        self.assertTrue(isinstance(wrapper['item1'], WrapperTest))
        self.assertTrue(isinstance(wrapper['item2'], WrapperTest))
        self.assertEqual(2, wrapper['item1'].item2.value)

    def test_kwargs(self):
        def func(item1, item2):
            self.assertTrue(isinstance(item1, WrapperTest))
            self.assertTrue(isinstance(item2, WrapperTest))

        wrapped = {'item1': Aux(), 'item2': Aux()}
        wrapper = WrapperTest.wrap_kwargs(wrapped)
        func(**wrapper)

    def test_getattr(self):
        wrapped = str('value')
        name = WrapperTest.will_wrap(wrapped)
        obj = Item(10)
        self.assertEqual(10, obj.value)
        getattr(obj, 'value')
        getattr(obj, name)

    def test_wrap_simple(self):
        def func():
            wrapped = str('value')
            WrapperTest(wrapped)

        self.assertRaises(TypeError, func)

    def test_func_wrapper(self):
        def func():
            return 10

        wrapper = WrapperTest(func)
        res = wrapper()
        self.assertEqual(10, res)

    def test_set(self):
        wrapped = {'item1', 'item2'}
        wrapper = WrapperTest(wrapped)
        for s1, s2 in zip(wrapped, wrapper):
            self.assertEqual(s1, s2)

    def test_dict_items(self):
        wrapped = {'item1': ['test1', 'test2'], 'item2': ['test3', 'test4']}
        wrapper = WrapperTest.will_wrap(wrapped)
        for k, v in wrapper.items():
            self.assertTrue(isinstance(v, WrapperTest))

    def test_signature_join(self):
        a = 10
        b = 'test'
        c = Aux()
        result = WrapperData.signature_join(a, b, c)
        self.assertEqual('10 test Aux', result)

    def test_get_signature_from_obj(self):
        a = None
        result = WrapperData.get_signature_from_obj(a)
        self.assertIsNone(result)
        a = 10
        result = WrapperData.get_signature_from_obj(a)
        self.assertEqual('10', result)
        a = Aux()
        result = WrapperData.get_signature_from_obj(a)
        self.assertEqual('Aux', result)


if __name__ == '__main__':
    unittest.main()



from abc import ABC

import wrapt
import copy

from wafec_wrapt_custom.managers import DataEvent, DataManagerAdapter

_simple_types = (str, int, type(None), bool, float)


class WrapperTest(wrapt.ObjectProxy, ABC):
    def __init__(self, wrapped, wrapper_data=None):
        super(WrapperTest, self).__init__(WrapperTest.unwrap(wrapped))
        self._validate_type()
        self._self_wrapper_data = wrapper_data

    def _validate_type(self):
        if isinstance(self.__wrapped__, _simple_types):
            raise TypeError('Cannot be simple')

    def __getitem__(self, item):
        value = super(WrapperTest, self).__getitem__(item)
        wrapper_data = self.notify_external(item, 'getitem', value)
        return WrapperTest.will_wrap(value, wrapper_data=wrapper_data)

    def __setitem__(self, item, value):
        if isinstance(value, WrapperTest):
            super(WrapperTest, self).__setitem__(item, value.__wrapped__)
        else:
            super(WrapperTest, self).__setitem__(item, value)

    def __getattr__(self, name):
        value = super(WrapperTest, self).__getattr__(name)
        wrapper_data = self.notify_external(name, 'getattr', value)
        return WrapperTest.will_wrap(value, wrapper_data=wrapper_data)

    def __setattr__(self, name, value):
        if isinstance(value, WrapperTest):
            super(WrapperTest, self).__setattr__(name, value.__wrapped__)
        else:
            super(WrapperTest, self).__setattr__(name, value)

    def __copy__(self):
        return WrapperTest.will_wrap(copy.copy(self.__wrapped__))

    def __deepcopy__(self, memodict={}):
        return WrapperTest.will_wrap(copy.deepcopy(self.__wrapped__, memodict))

    def __iter__(self):
        self._self_iter_idx = 0
        self._self_iter = iter(self.__wrapped__)
        return self

    def __next__(self):
        value = next(self._self_iter)
        wrapper_data = self.notify_external(self._self_iter_idx, 'iter', value)
        self._self_iter_idx += 1
        return WrapperTest.will_wrap(value, wrapper_data=wrapper_data)

    def notify_external(self, signature, event_name, value):
        is_callable = False
        data_type = 'undefined'
        if value is not None:
            if isinstance(value, WrapperTest):
                value = value.__wrapped__
            is_callable = callable(value)
            data_type = type(value)

        if self._self_wrapper_data:
            wrapper_data = WrapperData.of(signature, self._self_wrapper_data)
            self._self_wrapper_data.manager.notify(DataEvent(wrapper_data.signature, event_name, value, data_type,
                                                             is_callable))
            return wrapper_data
        return self._self_wrapper_data

    @staticmethod
    def wrap_kwargs(kwargs, wrapper_data=None):
        for argname, argvalue in kwargs.items():
            kwargs[argname] = WrapperTest.will_wrap(argvalue, wrapper_data=WrapperData.of(argname, wrapper_data))
        return kwargs

    @staticmethod
    def unwrap(obj):
        if obj is not None and isinstance(obj, WrapperTest):
            return obj.__wrapped__
        return obj

    @staticmethod
    def will_wrap(obj, wrapper_data=None):
        if isinstance(obj, _simple_types):
            return obj
        return WrapperTest(obj, wrapper_data=wrapper_data)

    def __call__(self, *args, **kwargs):
        value = self.__wrapped__(*args, **kwargs)
        wrapper_data = self.notify_external(None, 'return', value)
        return WrapperTest.will_wrap(value, wrapper_data=wrapper_data)


class WrapperData(object):
    def __init__(self, manager, signature):
        self.manager = manager
        self.signature = signature

    @staticmethod
    def of(signature, wrapper_data):
        if not wrapper_data:
            return WrapperData(DataManagerAdapter(), WrapperData.get_signature_from_obj(signature, '?'))
        if wrapper_data.signature is not None:
            if signature is not None:
                signature = WrapperData.signature_join(wrapper_data.signature, signature)
            else:
                signature = WrapperData.get_signature_from_obj(wrapper_data.signature)
        else:
            signature = WrapperData.get_signature_from_obj(signature, '?')
        manager = wrapper_data.manager if wrapper_data.manager is not None else DataManagerAdapter()
        return WrapperData(manager, signature)

    @staticmethod
    def get_signature_from_obj(obj, default=None):
        signature = default
        if obj is not None:
            if isinstance(obj, _simple_types):
                signature = str(obj)
            else:
                signature = type(obj).__name__
        return signature

    @staticmethod
    def signature_join(*args):
        signature_list = []
        for arg in args:
            if arg is not None:
                signature_list.append(WrapperData.get_signature_from_obj(arg))
        if len(signature_list) == 0:
            return None
        else:
            return ' '.join(signature_list)


from setuptools import setup, find_packages


setup(
    name="wafec.wrapt.custom",
    version="1.0.4",
    author="Wallace",
    author_email="wallacefcardoso@gmail.com",
    packages=find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent"
    ],
    install_requires=[
        'wrapt>=1.12.1',
        'requests>=2.23.0'
    ]
)
